#---------------------------------------------------
# mtzviewer.tcl
#---------------------------------------------------
#
# Front end for MTZ viewer/editor application
#
#Cvs_Id $Id: mtzviewer.tcl,v 1.15 2006/10/11 00:42:25 pjx Exp $
#---------------------------------------------------

#################################################################
# Generic code starts
#
# The following code is common to any standalone application
# that uses the CCP4i core code
# This could be farmed out to a library or similar?
#################################################################

# Hide the Tk window
if { [lsearch -exact [package names] tk] >= -1 } {
    wm withdraw .
}
#
#----------------------------------------------------------------
proc GetEnvPath { var } {
#----------------------------------------------------------------
  global system
  global env

  # Get environment variable - cope with whether input var has
  # leading '$' or not

  regsub {^\$} $var  {} var1

  # On NT change the separators to unix style /

  if { [regexp WINDOWS $system(OPSYS)] } {
    set status [catch { global env;
      regsub -all {\\} $env($var1) \/ path } nc ]
  } else {
    set status [catch {global env; set p1 $env($var1)} path ]
  }
  if { $status == 1 } {
    set path ""
    WarningMessage "Can not get environment variable for $var"
  }
  return $path
}

# Source startup.tcl appropriate for operating system

set system(OPSYS) [string toupper $tcl_platform(platform)]
source [file join $env(CCP4I_TOP) bin $system(OPSYS) startup.tcl]

# Edit the RUN_MODE to be the name of the current application
set system(RUN_MODE)            mtzviewer
set system(TK_VERSION)          $tk_version

source [file join $env(CCP4I_TOP) src system.tcl]
source [file join $env(CCP4I_TOP) src utils.tcl]

# If there is an autoconfigure procedure for the system parameters
# and this opsys then run it
if { [llength [info proc autoconf_[subst $system(OPSYS)]_system]] > 0 } {
      autoconf_[subst $system(OPSYS)]_system system }

InitialiseDotCCP4
global configure
InitialisePreferences configure configure
InitialisePreferences preferences preferences
InitialisePreferences directories directories

#################################################################
# End of generic code
#
# Application specific code starts below here
#################################################################

# Show the Tk window
wm deiconify .

# Source files with hierarchical viewer functions
source [SearchPath TOP utils hierarch_utils.tcl]

# CCP4i core functions
source [SearchPath TOP src CCP4_utils.tcl]

source [SearchPath TOP src utils.tcl]
source [SearchPath TOP src window.tcl]
source [SearchPath TOP src exframe.tcl]
source [SearchPath TOP src varmenu.tcl]
source [SearchPath TOP src fileselect.tcl]
source [SearchPath TOP src fileviewer.tcl]
source [SearchPath TOP src util_windows.tcl]
source [SearchPath TOP etc types.def]


#################################################################
# Global variables
#################################################################

# LOADED indicates whether a data structure is
# currently loaded/displayed
set mtzviewer(LOADED) 0

# DEFAULT_INFO is the default text to show in the
# information bar
set mtzviewer(DEFAULT_INFO) ""

# COMMAND_BUFFER is the list of commands that will
# be sent to the MTZman program to effect the edits
set mtzviewer(COMMAND_BUFFER) {}

# SOURCE_MTZ is the original file loaded into the
# editor at the start
set mtzviewer(SOURCE_MTZ) ""

# UNSAVED_EDITS indicates whether there are editing
# operations that have been performed since the last save
set mtzviewer(UNSAVED_EDITS) 0

# NEW_CELL is used to collect the user-entered cell
# parameters required when making a new crystal
set mtzviewer(NEW_CELL) {}

#################################################################
# Application procedures
#################################################################

#----------------------------------------------------------------
proc buffer_command { cmd } {
#----------------------------------------------------------------
    # Store a script commands to be held in a buffer
    # When the user decides to save the MTZ structure to file,
    # these will be fed to the appropriate program to do the
    # work in background
    global mtzviewer
    lappend mtzviewer(COMMAND_BUFFER) $cmd
    # Also echo to the screen for debugging
    puts "Buffering command: $cmd"
    # Flag that there are unsaved edits
    set mtzviewer(UNSAVED_EDITS) 1
}

#----------------------------------------------------------------
proc dump_command_buffer { } {
#----------------------------------------------------------------
    # Retrieve the script commands held in the buffer
    global mtzviewer
    set command_buffer ""
    foreach cmd $mtzviewer(COMMAND_BUFFER) {
	append command_buffer $cmd "\n"
    }
    return [string trim $command_buffer "\n"]
}

#----------------------------------------------------------------
proc clear_command_buffer { } {
#----------------------------------------------------------------
    # Reset the buffer, removing all previously stored
    # script commands
    # Also reset the unsaved edits flag
    global mtzviewer
    set mtzviewer(COMMAND_BUFFER) {}
    set mtzviewer(UNSAVED_EDITS) 0
    return
}

#----------------------------------------------------------------
proc unsaved_edits { } {
#----------------------------------------------------------------
    # Return whether there are unsaved edits in the buffer
    global mtzviewer
    return $mtzviewer(UNSAVED_EDITS)
}

#----------------------------------------------------------------
proc enter_cell { } {
#----------------------------------------------------------------
    # Provide the user with a window to enter cell parameters

    set w .entercell
    if ![OpenWindow $w "Enter Crystal Cell Parameters" "Cell Parameters" -message ] {
	return
    }

    # Make the frame and the exit, add and apply buttons
    CreateFrame  $w entercell
    ##CreateButton $w dismiss "Exit"  "unset entercell ; destroy $w"
    CreateButton $w apply   "Ok" "apply_cell $w entercell"

    # Set the parameters to be edited
    DefineVariable entercell CELL_1 _cell_length ""
    DefineVariable entercell CELL_2 _cell_length ""
    DefineVariable entercell CELL_3 _cell_length ""
    DefineVariable entercell CELL_4 _cell_angle ""
    DefineVariable entercell CELL_5 _cell_angle ""
    DefineVariable entercell CELL_6 _cell_angle ""

    # Draw the widgets in the window
    CreateLine line \
	label "Enter cell parameters for new crystal:" -italic

    CreateLine line \
	label "Cell a:" \
	widget CELL_1 \
	label "b:" \
	widget CELL_2 \
	label "c:" \
	widget CELL_3 \
	label " (Alpha:" \
	widget CELL_4 \
	label "Beta:" \
	widget CELL_5 \
	label "Gamma:" \
	widget CELL_6 \
	label ")"

    # Must always do this at the end
    CloseFrame
    return
}

#----------------------------------------------------------------
proc apply_cell { w entercellVar } {
#----------------------------------------------------------------
    # Check and apply the cell parameters supplied by the user
    upvar #0 $entercellVar entercell
    global mtzviewer

    set cell {}

    # Check the cell lengths
    set ierr 0
    for { set i 1 } { $i <= 3 } { incr i } {
	if { ![regexp -- "^\[0-9\]+\.?\[0-9\]*\$" \
		   [string trim $entercell(CELL_$i)]] } {
	    set ierr 1
	} else {
	    lappend cell $entercell(CELL_$i)
	}
    }
    # Check the cell angles
    for { set i 4 } { $i <= 6 } { incr i } {
	if { [string trim $entercell(CELL_$i)] != "" } {
	    if { ![regexp -- "^\[0-9\]+\.?\[0-9\]*\$" \
		       [string trim $entercell(CELL_$i)]] } {
		set ierr 1
	    } else {
		lappend cell $entercell(CELL_$i)
	    }
	} else {
	    # Blank so assume it's 90 degrees
	    lappend cell 90.0
	}
    }
    # If there is an error then tell the user and send them
    # back again
    if { $ierr } {
	WarningMessage "Invalid cell parameters"
	return
    }
    # Close the temporary window
    unset entercell
    destroy $w
    # Update the cell
    set mtzviewer(NEW_CELL) $cell
    return
}

#----------------------------------------------------------------
proc load_mtzfile { { mtzfile "" } } {
#----------------------------------------------------------------
    # Delete existing view and load a new file in
    global mtzviewer
    
    # Query the user - do they want to do this?
    if { $mtzviewer(LOADED) } {
	set action [ChooseOptionDialog "Delete Existing MTZ Data" "Load MTZ" \
			"Loading a new file will delete the existing display.\n\nAre you sure you wish to do this?" \
			[list Yes No] \
			-default 1 ]
	if { $action == "No" } {
	    return
	}
    }
    # Delete the top-level node and reset globals
    hierarch_delete_all_nodes .top.vport.frame.fdisp
    set mtzviewer(LOADED) 0
    set mtzviewer(DEFAULT_INFO) ""
    set mtzviewer(SOURCE_MTZ) ""
    clear_xtal_info

    # Clear any outstanding edits
    clear_command_buffer

    # Import the new file
    set mtzviewer(SOURCE_MTZ) [import_mtzfile -mtzfile "$mtzfile" -nocommand]
}

#----------------------------------------------------------------
proc import_mtzfile { args } {
#----------------------------------------------------------------
    # This adds the contents of a new file into the existing
    # display, and returns the name of the imported file
    # Arguments are:
    # -mtzfile <filename>: load this file rather than asking
    # the user to select via the file browser
    # -nocommand: don't write an IMPORT command
    global mtzviewer

    # Process arguments
    set write_command 1
    if { [lsearch $args "-nocommand"] > -1 } {
	set write_command 0
    }
    set mtzfile ""
    if { [set i [lsearch $args "-mtzfile"]] > -1 } {
	incr i
	set mtzfile [lindex $args $i]
    }

    # Ask user for a file if one wasn't specified
    if { "$mtzfile" == "" } {
	set status [SelectFile mtzfile "Mtz file" -filter "*.mtz"]
    }
    
    # File exists will be successful even if the "file" is
    # actually a directory, so check
    if { [file exists $mtzfile ] && [file isfile $mtzfile] } {
	display_mtz $mtzfile
	set mtzviewer(DEFAULT_INFO) "File: $mtzfile"
	clear_xtal_info
	# Add the import command to the buffer
	if { $write_command } {
	    buffer_command "IMPORT $mtzfile"
	}
	# Return the filename
	return $mtzfile
    }
}

#----------------------------------------------------------------
proc display_mtz { mtzfile } {
#----------------------------------------------------------------
    # Create the hierarchical view of the MTZ file
    global mtzviewer

    if { ![file exists $mtzfile] } {
	return
    }

    set vport .top.vport
    set fdisplay $vport.frame.fdisp

    # Descend the MTZ hierarchy and build the graphical representation
    foreach xtal [GetMtzXtals $mtzfile] {

	# Add crystals
	if { [set fxtal [add_xtal $xtal -nocommand -noupdate]] != "" } {
	    hierarch_set_node_data_item $fxtal "NAME" $xtal
	    hierarch_set_node_data_item $fxtal "CELL" [get_xtal_cell $mtzfile $xtal]

	    # Add datasets
	    foreach dataset [GetMtzDatasetsInXtal $mtzfile $xtal] {
		set fdata [add_dataset $fxtal $dataset -nocommand -noupdate]
		hierarch_set_node_data_item $fdata "NAME" $dataset
		hierarch_set_node_data_item $fdata "WAVE" \
		    [get_dataset_wave $mtzfile $xtal $dataset]

		# Add columns
		foreach col [GetMtzColsInDataset $mtzfile $xtal $dataset] {
		    set fcol [add_column $fdata $col -noupdate]
		    hierarch_set_node_data_item $fcol "NAME" $col
		    hierarch_set_node_data_item $fcol "TYPE" \
			[GetMtzColType $mtzfile $col]
		}
	    }
	} else {
	    WarningMessage "Failed to import crystal\n\"$xtal\"\nfrom file"
	}
    }
    
    # Update the display - forces drawing of connecting lines
    hierarch_update_display $fdisplay

    set mtzviewer(LOADED) 1
}

#----------------------------------------------------------------
proc resize_view { top } {
#----------------------------------------------------------------
    # Reset the viewable area of the canvas whenever the size
    # of the hierarchical view changes
    set vport $top.vport
    # Maximum height etc are set arbitrarily here
    set maxheight 600
    set minheight 250
    set minwidth 350
    # Reset the maximum scrollable region
    set bbox [$vport bbox all]
    set height [winfo height $vport.frame]
    if { $height > $maxheight } {
	set height $maxheight
    } elseif { $height < $minheight } {
	set height $minheight
    }
    set width [winfo width $vport.frame]
    if { $width < $minwidth } {
	set width $minwidth
    }
    $vport configure -width $width -height $height -scrollregion $bbox
}

#----------------------------------------------------------------
proc click_on_xtal { w x y } {
#----------------------------------------------------------------
    # Invoked when the user right-clicks on a crystal node
    # Pops up a menu with various options for editing
    # Make a pop-up menu
    set m "$w.edit"
    if { [winfo exists $m] } {
	destroy $m
    }
    menu $m -tearoff 0
    # Add the context-specific options first
    $m add command -label "Edit name" -command "hierarch_edit_node_name $w check_xtal_name"
    $m add command -label "Delete crystal [hierarch_get_node_name $w]" \
	-command "delete_xtal $w"
    $m add command -label "Add dataset" -command "add_dataset $w"
    # Add the standard options at the end
    append_standard_options $m

    tk_popup $m $x $y
}

#----------------------------------------------------------------
proc click_on_dataset { w x y } {
#----------------------------------------------------------------
    # Invoked when the user right-clicks on a dataset node
    set m "$w.edit"
    if { [winfo exists $m] } {
	destroy $m
    }
    menu $m -tearoff 0
    # Add the context-specific options first
    $m add command -label "Edit name" -command "hierarch_edit_node_name $w check_dataset_name"
    $m add command -label "Delete dataset [hierarch_get_node_name $w]" \
	-command "delete_dataset $w"

    # Option to move the dataset to another crystal
    set parent_xtal [hierarch_get_parent $w]
    set xtals [list_xtals]
    if { [llength $xtals] > 1 } {
	# Make a cascading menu of possible crystals
	set mc $m.move
	if { [winfo exists $mc] } {
	    destroy $mc
	}
	menu $mc -tearoff 0
	foreach xtal $xtals {
	    if { $xtal != $parent_xtal } {
		$mc add command -label "[hierarch_get_node_name $xtal]" \
		    -command "move_dataset $w $xtal"
	    }
	}
	$m add cascade -label "Move dataset to.." -menu $mc
    }
    
    # Add the standard options at the end
    #append_standard_options $m

    tk_popup $m $x $y
}

#----------------------------------------------------------------
proc click_on_column { w x y } {
#----------------------------------------------------------------
    # Invoked when the user right-clicks on a column node
    set m "$w.edit"
    if { [winfo exists $m] } {
	destroy $m
    }
    menu $m -tearoff 0
    # Add the context-specific options first
    $m add command -label "Edit name" -command "hierarch_edit_node_name $w check_column_name"
    $m add command -label "Delete column [hierarch_get_node_name $w]" \
	-command "delete_column $w"

    # Option to move the column to another crystal/dataset pair
    set parent_dataset [hierarch_get_parent $w]
    set parent_xtal [hierarch_get_parent $parent_dataset]
    set xtals [list_xtals]
    if { [llength $xtals] > 1 } {
	# Make a cascading menu of possible crystal/dataset pairs
	set mc $m.move
	if { [winfo exists $mc] } {
	    destroy $mc
	}
	menu $mc -tearoff 0
	foreach xtal $xtals {
	    foreach dataset [list_datasets $xtal] {
		if { $xtal == $parent_xtal && $dataset == $parent_dataset } {
		    # Skip this one
		} else {
		    $mc add command -label "/[hierarch_get_node_name $xtal]/[hierarch_get_node_name $dataset]" \
			-command "move_column $w $xtal $dataset"
		}
	    }
	}
	# Add the completed cascading menu
	$m add cascade -label "Move column to.." -menu $mc
    }
    
    # Add the standard options at the end
    #append_standard_options $m

    tk_popup $m $x $y
}

#----------------------------------------------------------------
proc check_xtal_name { w name } {
#----------------------------------------------------------------
    # Invoked when user edits the name of a crystal
    # No blank names
    if { [string trim $name] == "" } {
	return 0
    }
    # Check that it only contains allowed characters
    if { ![regexp -- "^\[A-Za-z0-9_\]+$" $name] } {
	WarningMessage "The crystal name:

\"$name\"

is not valid.

Crystal names can only contain
alphanumeric characters and the
underscore character."
	return 0
    }
    # Check the length
    if { [string length $name] > 64 } {
	WarningMessage "The crystal name:

\"$name\"

is too long - names cannot be
longer than 64 characters."
	return 0
    }
    # Check for uniqueness
    if { [xname_exists $name] } {
	WarningMessage "A crystal with the name

$name

already exists"
	return -1
    }
    # Write the command to the buffer
    buffer_command "RENAME XTAL [hierarch_get_node_name $w] $name"
    return 1
}

#----------------------------------------------------------------
proc check_dataset_name { w name } {
#----------------------------------------------------------------
    # Invoked when user edits the name of a dataset
    # No blank names
    if { [string trim $name] == "" } {
	return 0
    }
    # Check that it only contains allowed characters
    if { ![regexp -- "^\[A-Za-z0-9_\]+$" $name] } {
	WarningMessage "The dataset name:

\"$name\"

is not valid.

Dataset names can only contain
alphanumeric characters and the
underscore character."
	return 0
    }
    # Check the length
    if { [string length $name] > 64 } {
	WarningMessage "The dataset name:

\"$name\"

is too long - names cannot be
longer than 64 characters."
	return 0
    }
    # Check for uniqueness
    set wxtal [hierarch_get_parent $w]
    if { [dname_exists $name $wxtal] } {
	WarningMessage "A dataset with the name

$name

already exists"
	return -1
    }
    # Write the command to the buffer
    buffer_command "RENAME DATASET /[hierarch_get_node_name [hierarch_get_parent $w]]/[hierarch_get_node_name $w] $name"
    return 1
}

#----------------------------------------------------------------
proc check_column_name { w name } {
#----------------------------------------------------------------
    # Invoked when user edits the name of a column
    # No blank names
    if { [string trim $name] == "" } {
	return 0
    }
    # Check that it only contains allowed characters
    if { ![regexp -- "^\[A-Za-z0-9_\]+$" $name] } {
	WarningMessage "The dataset name:

\"$name\"

is not valid.

Column names can only contain
alphanumeric characters and the
underscore character."
	return 0
    }
    # Check the length
    if { [string length $name] > 64 } {
	WarningMessage "The column name:

\"$name\"

is too long - names cannot be
longer than 64 characters."
	return 0
    }
    # Check for uniqueness
    set wdata [hierarch_get_parent $w]
    if { [colname_exists $name] } {
	WarningMessage "A column with the name

$name

already exists."
	return -1
    }
    # Write the command to the buffer
    set wdata [hierarch_get_parent $w]
    set dname [hierarch_get_node_name $wdata]
    set wxtal [hierarch_get_parent $wdata]
    set xname [hierarch_get_node_name $wxtal]
    buffer_command "RENAME COLUMN /$xname/$dname/[hierarch_get_node_name $w] /$xname/$dname/$name"
    return 1
}

#----------------------------------------------------------------
proc show_xtal_info { w x y } {
#----------------------------------------------------------------
    # Invoked when the cursor enters a crystal node
    # Put the name of the crystal in the message bar
    global mtzviewer
    set info $mtzviewer(INFORMATION_BAR)
    set xtal_name [hierarch_get_node_name $w]
    set xtal_cell [hierarch_get_node_data_item $w "CELL"]
    $info configure -text "Crystal: \"$xtal_name\"\nCell: $xtal_cell"
}

#----------------------------------------------------------------
proc clear_xtal_info { { w "" } { x "" } { y "" } } {
#----------------------------------------------------------------
    # Invoked when the cursor leaves a crystal node
    # Clear the information previously displayed
    # FIXME could be consolidated with clear_dataset_info?
    global mtzviewer
    set info $mtzviewer(INFORMATION_BAR)
    set xtal_name [hierarch_get_node_name $w]
    $info configure -text "$mtzviewer(DEFAULT_INFO)"
}

#----------------------------------------------------------------
proc show_dataset_info { w x y } {
#----------------------------------------------------------------
    # Invoked when the cursor enters a dataset node
    # Put the name of the dataset in the message bar
    global mtzviewer
    set info $mtzviewer(INFORMATION_BAR)
    set dataset_name [hierarch_get_node_name $w]
    set dataset_wave [hierarch_get_node_data_item $w "WAVE"]
    $info configure -text "Dataset: \"$dataset_name\"\nWavelength: $dataset_wave (A)"
}

#----------------------------------------------------------------
proc clear_dataset_info { { w "" } { x "" } { y "" } } {
#----------------------------------------------------------------
    # Invoked when the cursor leaves a dataset node
    # Clear the information previously displayed
    # FIXME could be consolidated with clear_xtal_info?
    global mtzviewer
    set info $mtzviewer(INFORMATION_BAR)
    set xtal_name [hierarch_get_node_name $w]
    $info configure -text "$mtzviewer(DEFAULT_INFO)"
}

#----------------------------------------------------------------
proc show_column_info { w x y } {
#----------------------------------------------------------------
    # Invoked when the cursor enters a column node
    # Put the name of the column in the message bar
    global mtzviewer
    set info $mtzviewer(INFORMATION_BAR)
    set column_name [hierarch_get_node_name $w]
    set column_type [hierarch_get_node_data_item $w "TYPE"]
    $info configure -text "Column: \"$column_name\"\tType: $column_type"
}

#----------------------------------------------------------------
proc clear_column_info { { w "" } { x "" } { y "" } } {
#----------------------------------------------------------------
    # Invoked when the cursor leaves a column node
    # Clear the information previously displayed
    # FIXME could be consolidated with clear_xtal_info?
    global mtzviewer
    set info $mtzviewer(INFORMATION_BAR)
    set xtal_name [hierarch_get_node_name $w]
    $info configure -text "$mtzviewer(DEFAULT_INFO)"
}

#----------------------------------------------------------------
proc get_xtal_cell { mtzfile xtal } {
#----------------------------------------------------------------
    # Return the cell parameters for a crystal in the named
    # MTZ file
    set cell {}
    set datasets [GetMtzDatasetsInXtal $mtzfile $xtal]
    if { [llength $datasets] > 0 } {
	set dataset [lindex $datasets 0]
	for { set i 1 } { $i<=6 } { incr i } {
	    if { [GetMtzParamFromDataset \
		       $mtzfile $xtal $dataset DCELL_$i dcell] } {
		##puts "get_xtal_cell: DCELL_$i returned as $dcell"
		lappend cell $dcell
	    }
	}
    }
    if { [llength $cell] == 6 } {
	# Correct number of cell parameters
	return $cell
    } else {
	# Failed to acquire correct number of
	# cell parameters
	return [list "?" "?" "?" "?" "?" "?"]
    }
}

#----------------------------------------------------------------
proc get_dataset_wave { mtzfile xtal dataset } {
#----------------------------------------------------------------
    # Return the wavelength for the dataset in the specified xtal
    # and MTZ file
    set wave ""
    if { [GetMtzParamFromDataset $mtzfile $xtal $dataset DWAVES dwaves] } {
	##puts "get_dataset_wave: DWAVES returned as $dwaves"
	set wave $dwaves
    }
    return $wave
}

#----------------------------------------------------------------
proc append_standard_options { m } {
#----------------------------------------------------------------
    # Append a set of standard options to a pop-up menu
    $m add command -label "New crystal" -command "add_xtal"
    # Check if anything is selected
    # and offer to delete it if there is
    set fdisplay .top.vport.frame.fdisp
    set selection [return_selected_nodes $fdisplay]
    if { [llength $selection] > 0 } {
	$m add command -label "Delete selected items" \
	    -command "delete_selected \"$selection\""
    }
}

#----------------------------------------------------------------
proc add_xtal { { name "" } args } {
#----------------------------------------------------------------
    # Add a new crystal to the current view
    # Arguments are:
    # -noupdate: same as for the hierarch_add_node command
    # -nocommand: don't write a NEW XTAL command
    global mtzviewer

    # Process arguments
    set update 1
    set write_command 1
    if { [lsearch $args "-noupdate"] > -1 } {
	set update 0
    }
    if { [lsearch $args "-nocommand"] > -1 } {
	set write_command 0
    }
    
    # Fetch name of root node widget
    set fdisplay .top.vport.frame.fdisp
    if { $name == "" } {
	set base_name "new_xtal"
    } else {
	set base_name $name
    }
    set new_xtal $base_name 
    set i -1
    while { [xname_exists $new_xtal] > 0 } {
	puts "WARNING crystal $new_xtal already exists"
	incr i
	if { $i == 10 } {
	    puts "ERROR failed to make a unique name, quitting"
	    return ""
	}
	set new_xtal "[subst $base_name]$i"
    }
    set add_cmd "hierarch_add_node $fdisplay $new_xtal -callback click_on_xtal -mouse_enter show_xtal_info -mouse_leave clear_xtal_info"
    if { ! $update } {
	append add_cmd " -noupdate"
    }
    set fxtal [eval $add_cmd]
    if { $write_command } {
	# User has created a new crystal
	# Get the cell parameters
	enter_cell
	vwait mtzviewer(NEW_CELL)
	# Update the associated data
	hierarch_set_node_data_item $fxtal "CELL" $mtzviewer(NEW_CELL)
	# Add to the command buffer
	buffer_command "NEW XTAL $new_xtal CELL $mtzviewer(NEW_CELL)"
	# Automatically set this into editing mode
	hierarch_edit_node_name $fxtal check_xtal_name
    }
    return $fxtal
}

#----------------------------------------------------------------
proc add_dataset { w { name "" } args } {
#----------------------------------------------------------------
    # Add a new dataset to the current view, under the crystal
    # at node w
    # Arguments are:
    # -noupdate: same as for the hierarch_add_node command
    # -nocommand: don't write a NEW DATASET command

    # Process arguments
    set update 1
    set write_command 1
    if { [lsearch $args "-noupdate"] > -1 } {
	set update 0
    }
    if { [lsearch $args "-nocommand"] > -1 } {
	set write_command 0
    }

    # Set the dataset name
    # FIXME need to impose uniqueness restraint here
    if { "$name" == "" } {
	set base_name "new_dataset"
    } else {
	set base_name $name
    }
    set new_dataset $base_name 
    set i -1
    while { [dname_exists $new_dataset $w] > 0 } {
	puts "WARNING dataset $new_dataset already exists in this xtal"
	incr i
	if { $i == 10 } {
	    puts "ERROR failed to make a unique name, quitting"
	    return ""
	}
	set new_dataset "[subst $base_name]$i"
    }

    # Setup the command to add the dataset
    set add_cmd "hierarch_add_node $w $new_dataset -callback click_on_dataset -mouse_enter show_dataset_info -mouse_leave clear_dataset_info"
    if { ! $update } {
	append add_cmd " -noupdate"
    }
    set fdata [eval $add_cmd]
    if { $write_command } {
	# User has created a new dataset so add to the
	# command buffer
	buffer_command "NEW DATASET /[hierarch_get_node_name $w] $new_dataset"
	# FIXME Automatically set this into editing mode
	hierarch_edit_node_name $fdata check_dataset_name
    }
    return $fdata
}

#----------------------------------------------------------------
proc add_column { w name args } {
#----------------------------------------------------------------
    # Add a new column to the current view, under the dataset
    # at node w
    # Arguments are:
    # -noupdate: same as for the hierarch_add_node command

    # Process arguments
    set update 1
    set write_command 1
    if { [lsearch $args "-noupdate"] > -1 } {
	set update 0
    }

    # Set the column name
    # FIXME need to impose uniqueness restraint here
    set base_name $name
    set new_column $base_name
    set i -1
    while { [colname_exists $new_column] > 0 } {
	puts "WARNING column called \"$new_column\" already exists"
	incr i
	if { $i == 10 } {
	    puts "ERROR failed to make a unique name, quitting"
	    return ""
	}
	set new_column "[subst $base_name]$i"
    }

    # Setup the command to add the column
    set add_cmd "hierarch_add_node $w $new_column -callback click_on_column -mouse_enter show_column_info -mouse_leave clear_column_info -nochildren"
    if { ! $update } {
	append add_cmd " -noupdate"
    }
    set fcol [eval $add_cmd]
    return $fcol
}

#----------------------------------------------------------------
proc delete_xtal { w } {
#----------------------------------------------------------------
    # Delete the specified (crystal) node
    puts "Deleting $w and its children"
    buffer_command "DELETE XTAL [hierarch_get_node_name $w]"
    hierarch_delete_node $w
}

#----------------------------------------------------------------
proc delete_dataset { w args } {
#----------------------------------------------------------------
    # Delete the specified (dataset) node
    puts "Deleting $w and its children"
    buffer_command "DELETE DATASET /[hierarch_get_node_name [hierarch_get_parent $w]]/[hierarch_get_node_name $w]"
    hierarch_delete_node $w
    # Force redrawing of connecting lines
    set fdisplay .top.vport.frame.fdisp
    hierarch_update_display $fdisplay
}

#----------------------------------------------------------------
proc delete_column { w args } {
#----------------------------------------------------------------
    # Delete the specified (column) node
    puts "Deleting $w"
    set wdata [hierarch_get_parent $w]
    set dname [hierarch_get_node_name $wdata]
    set wxtal [hierarch_get_parent $wdata]
    set xname [hierarch_get_node_name $wxtal]
    buffer_command "DELETE COLUMN /$xname/$dname/[hierarch_get_node_name $w]"
    hierarch_delete_node $w
    # Force redrawing of connecting lines
    set fdisplay .top.vport.frame.fdisp
    hierarch_update_display $fdisplay
}

#----------------------------------------------------------------
proc move_dataset { wdata wxtal } {
#----------------------------------------------------------------
    # Move a dataset from one parent crystal to another
    set wxtal_old [hierarch_get_parent $wdata]
    set old_xname [hierarch_get_node_name $wxtal_old]
    puts "Moving dataset [hierarch_get_node_name $wdata] from [hierarch_get_node_name $wxtal_old] to [hierarch_get_node_name $wxtal]"
    # Check that there isn't already a dataset with that name
    # in the target crystal
    if { [dname_exists [hierarch_get_node_name $wdata] $wxtal] } {
	WarningMessage "There is already a dataset with the
same name in the target crystal.

Please rename one of the datasets before
moving it."
	return
    }
    # "Move" is actually a copy operation followed by a delete
    set wdata_new [hierarch_copy_node $wdata $wxtal -all]
    hierarch_delete_node $wdata
    # Write the equivalent command to the buffer
    buffer_command "MOVE DATASET /$old_xname/[hierarch_get_node_name $wdata_new] /[hierarch_get_node_name $wxtal]"
    # Force redrawing of connecting lines
    set fdisplay .top.vport.frame.fdisp
    hierarch_update_display $fdisplay
    return $wdata_new
}

#----------------------------------------------------------------
proc move_column { wcol wxtal wdata } {
#----------------------------------------------------------------
    # Move a column from one parent crystal/dataset pair to another
    # Find the existing parents
    set wdata_old [hierarch_get_parent $wcol]
    set wxtal_old [hierarch_get_parent $wdata_old]
    # Get the names
    set old_xname [hierarch_get_node_name $wxtal_old]
    set old_dname [hierarch_get_node_name $wdata_old]
    puts "Moving column [hierarch_get_node_name $wcol] from /[hierarch_get_node_name $wxtal_old]/[hierarch_get_node_name $wdata_old] to /[hierarch_get_node_name $wxtal]/[hierarch_get_node_name $wdata]"
    # Check that there isn't already a column with that name
    # in the target crystal/dataset combination
    if { [colname_exists [hierarch_get_node_name $wcol] $wcol] } {
	WarningMessage "There is already a column with the
same name in the target crystal/dataset
pair.

Please rename one of the columns before
moving it."
	return
    }
    # "Move" is actually a copy operation followed by a delete
    set wcol_new [hierarch_copy_node $wcol $wdata -all]
    hierarch_delete_node $wcol
    # Write the equivalent command to the buffer
    buffer_command "MOVE COLUMN /$old_xname/$old_dname/[hierarch_get_node_name $wcol_new] /[hierarch_get_node_name $wxtal]/[hierarch_get_node_name $wdata]"
    # Force redrawing of connecting lines
    set fdisplay .top.vport.frame.fdisp
    hierarch_update_display $fdisplay
    return $wcol_new
}

#----------------------------------------------------------------
proc delete_selected { wnodes } {
#----------------------------------------------------------------
    # Delete a group of specified nodes all at once
    puts "$wnodes"
    foreach w $wnodes {
	delete_xtal $w
    }
}

#----------------------------------------------------------------
proc xname_exists { name } {
#----------------------------------------------------------------
    # See if a xtal node already exists with the given name
    set fdisplay .top.vport.frame.fdisp
    set wnodes [hierarch_find_node $fdisplay $name]
    ##puts "$name: $wnodes"
    if { [llength $wnodes] > 0 } {
	foreach w $wnodes {
	    set level [hierarch_node_level $w]
	    ##puts "$w: level = $level"
	    if { $level == 1 } {
		return 1
	    }
	}
    }
    return 0    
}

#----------------------------------------------------------------
proc dname_exists { name wxtal } {
#----------------------------------------------------------------
    # See if a dataset node already exists with the given name
    # Note that the dataset name is checked only in the parent
    # crystal (node wxtal)
    # FIXME should be consolidated with xname_exists
    set fdisplay .top.vport.frame.fdisp
    set wnodes [hierarch_find_node $fdisplay $name]
    ##puts "$name: $wnodes"
    if { [llength $wnodes] > 0 } {
	foreach w $wnodes {
	    set level [hierarch_node_level $w]
	    ##puts "$w: level = $level"
	    if { $level == 2 && [hierarch_get_parent $w] == $wxtal} {
		return 1
	    }
	}
    }
    return 0
}

#----------------------------------------------------------------
proc colname_exists { name { wcol "" } } {
#----------------------------------------------------------------
    # See if a column node already exists with the given name
    # The column label needs to be unique across the whole
    # MTZ data structure
    # FIXME should be consolidated with xname_exists and dname_exists?
    set fdisplay .top.vport.frame.fdisp
    set wnodes [hierarch_find_node $fdisplay $name]
    puts "colname_exists: name = $name"
    ##puts "$name: $wnodes"
    if { [llength $wnodes] > 0 } {
	foreach w $wnodes {
	    set level [hierarch_node_level $w]
	    ##puts "$w: level = $level"
	    if { $level == 3 } {
		# Only accept a match if it's not the same node
		if { $w != "$wcol" } {
		    return 1
		}
	    }
	}
    }
    return 0
}

#----------------------------------------------------------------
proc list_xtals { } {
#----------------------------------------------------------------
    # Return a list of crystal names
    set fdisplay .top.vport.frame.fdisp
    set xtals {}
    foreach w [hierarch_get_children $fdisplay] {
	lappend xtals $w
    }
    return $xtals
}

#----------------------------------------------------------------
proc list_datasets { fxtal } {
#----------------------------------------------------------------
    # Return a list of dataset names in the named crystal
    foreach w [hierarch_get_children $fxtal] {
	lappend datasets $w
    }
    return $datasets
}

#----------------------------------------------------------------
proc save_mtzfile { } {
#----------------------------------------------------------------
    # Save actually consists of running the MTZman program
    # on the original file in order to duplicate the operations
    # previously applied inside the editor
    global mtzviewer
    
    # Are there unsaved edits?
    if { ![unsaved_edits] } {
	# Nothing to save
	WarningMessage "No editing operations have been performed\nsince the last save"
	return
    }

    # First we need a file to write to
    set status [SelectFile new_mtzfile "Mtz file" -filter "*.mtz" -unknown]

    # Check that something was selected
    if { "$new_mtzfile" == "" || [file isdirectory $new_mtzfile] } {
	return
    }

    # If the file already exists then warn the user
    if { [file exists $new_mtzfile ] && [file isfile $new_mtzfile] } {
	set action [ChooseOptionDialog "Delete Existing MTZ File" "Save MTZ" \
			"The file\n\n\"$new_mtzfile\"\n\nalready exists - if you choose to continue then it will be overwritten.\n\nAre you sure you wish to do this?" \
			[list Yes No] \
			-default 1 ]
	if { $action == "No" } {
	    return
	}
    }
    puts "save_mtzfile: writing to file \"$new_mtzfile\""

    # Current source file
    set mtzfile "$mtzviewer(SOURCE_MTZ)"

    # Create the script to run the MTZman program
    set mtzman_cmd "cmtzman HKLIN $mtzfile HKLOUT $new_mtzfile"

    # Retrieve the script commands
    set mtzman_script [dump_command_buffer]

    # Run the commands
    puts "Run command:"
    puts "$mtzman_cmd <<EOF"
    puts "$mtzman_script"
    puts "EOF"

    WriteFile [set comfile [GetTmpFileName -ext _com ]] $mtzman_script
    if { [catch \
	      { exec cmtzman HKLIN $mtzfile HKLOUT $new_mtzfile < $comfile } \
	      errmsg ] } {
	# Caught an error
	puts "******** ERROR ********"
	puts $errmsg
	puts "********  END  ********"
    } else {
	# Successful outcome
	puts "No error from running cmtzman"
    }

    # Tidy up - clear the stored commands
    clear_command_buffer

    # Set this as the new current file
    set mtzviewer(SOURCE_MTZ) "$new_mtzfile"
    set mtzviewer(DEFAULT_INFO) "File: $new_mtzfile"
}

#----------------------------------------------------------------
proc init_mtzviewer { w } {
#----------------------------------------------------------------
    global configure
    global mtzviewer

    # Builds the visualiser window

    # Build a scrollable canvas
    set top [frame $w.top]

    set scroll [scrollbar $top.scroll -command "$top.vport yview"]
    pack $scroll -side right -fill y

    set vport [canvas $top.vport -yscrollcommand "$top.scroll set"]
    pack $vport -side left -fill both -expand true

    set frame [frame $vport.frame]
    $vport create window 0 0 -anchor nw -window $frame
    bind $frame <Configure> "resize_view $top"

    # Initialise the hierarchical viewer
    set fdisplay [hierarch_create $frame.fdisp ""]
    pack $fdisplay

    # Information bar
    set finfo [frame $w.finfo]
    set info  [label $finfo.info -text "" \
		   -height 3 \
		   -justify left -relief groove -anchor nw]
    pack $info -side left -expand 1 -fill both
    set mtzviewer(INFORMATION_BAR) $info

    #############################################################
    # Buttons and menu bar
    #############################################################
    set fmbar [frame .fmbar -background $configure(COLOUR_CONTRAST)]

    # File menu
    set bfile [ menubutton $fmbar.file -text "File" \
		    -background $configure(COLOUR_CONTRAST) \
		    -menu $fmbar.file.menu]
    set filemenu [ menu $bfile.menu -tearoff 0 ]
    $filemenu add command -label "Load new MTZ file" -command { load_mtzfile }
    $filemenu add command -label "Import from new MTZ file" -command { import_mtzfile }
    $filemenu add command -label "Save file" -command { save_mtzfile }
    $filemenu add command -label "Exit" -command { exit_mtzviewer }
    pack $bfile -side left

    # Crystal menu
    set bxtal [menubutton $fmbar.xtal -text "Crystal" \
		    -background $configure(COLOUR_CONTRAST) \
		    -menu $fmbar.xtal.menu]
    set xtalmenu [ menu $bxtal.menu -tearoff 0 ]
    $xtalmenu add command -label "Add new crystal" -command { add_xtal }
    pack $bxtal -side left

    # Pack everything
    pack $fmbar $top $finfo -side top -fill x
    return
}

#----------------------------------------------------------------
proc exit_mtzviewer { } {
#----------------------------------------------------------------
    global configure
    global mtzviewer

    # Check for unsaved changes
    if { [unsaved_edits] } {
	# Ask the user if they want to save their changes
	# before exiting
	set action [ChooseOptionDialog "Save changes before exiting" "Exit" \
			"You have made unsaved changes to the MTZ data\n\nDo you want to save these changes,\ndiscard them, or cancel the exit?" \
			[list Cancel Discard Save] \
			-default 0 ]
	if { $action == "Cancel" } {
	    # Don't exit
	    return
	} elseif { $action == "Save" } {
	    # Let them save their changes first
	    save_mtzfile
	}
    }
    # Exit
    exit
}

#################################################################
# Top-level script below here
#################################################################

# Read the command line arguments
set mtzfile ""
if { [llength $argv] == 1 } {
    set mtzfile [lindex $argv 0]
}

wm title . "CCP4Interface MtzViewer"

init_mtzviewer .

if { $mtzfile != "" } {
    load_mtzfile $mtzfile
}

