# ======================================================================
# phaker.tcl --
#
# Template for task interface
#
# CCP4Interface 
#
# =======================================================================

# Source files with hierarchical viewer functions
source [SearchPath TOP utils hierarch_utils.tcl]

#--------------------------------------------------------------------
proc phaker_setup { typedefname arrayname } { 
#--------------------------------------------------------------------

  upvar #0 $typedefname typedef

  # Different approach - define generic types for crystal and dataset
  set typedef(_mtz_xtal)    { mtz_label {*} Unassigned {*} }
  set typedef(_mtz_dataset) { mtz_label {*} Unassigned {*} }

  return 1
}

#--------------------------------------------------------------------
proc phaker_task_window { arrayname } {
#--------------------------------------------------------------------

  upvar #0 $arrayname array

  if { [CreateTaskWindow $arrayname  \
        "Phaker - Fake Task Interface" "Phaker" \
        [ list "Dataset Input"  ] ] == 0 } return

#=PROTOCOL==============================================================

  OpenFolder protocol 

  CreateTitleLine line TITLE

#=FILES================================================================

  OpenFolder file 

  CreateInputFileLine line \
        "Enter input MTZ file name (HKLIN)" \
        "MTZ in" \
        HKLIN DIR_HKLIN \
        -fileout HKLOUT DIR_HKLOUT _phaker

#---------------------------------------------------Datasets etc

  OpenFolder 1

  CreateLine line \
          message "View the data hierarchy in the MTZ file" \
          button "View crystal/dataset information" \
             -command "phaker_view_mtz $arrayname"

  # Repack the line so the button appears on the right
  # NB this seems to cause problems with closing/opening the folder
  pack forget $line.b1
  pack $line.b1 -side right

  CreateLine line label "Master crystal and dataset" -italic

  CreateMTZSelectLine line \
      "Select crystal and dataset from MTZ file" HKLIN \
      "Crystal:" XTALNAME \
      "Dataset:" DATASETNAME

  CreateLine line label "Additional crystal/datasets" -italic

  CreateExtendingFrame NDATASETS PhakerChooseDataset0 \
     "Select xtal-dataset pairs to be included" \
     "Add dataset" \
      [list RUN_XTAL RUN_DATASET]

  return
}

#-----------------------------------------------------------------------
proc phaker_run { arrayname } {
#-----------------------------------------------------------------------
# Write the xtal/dataset pairs to stdout
  upvar #0 $arrayname array

  puts "Phaker: master xtal/dataset = $array(XTALNAME)/$array(DATASETNAME)"

  # Return zero to preveny anything actually happening
  return 0
}

#-----------------------------------------------------------------------
proc PhakerChooseDataset0 { arrayname counter } {
#-----------------------------------------------------------------------
# Create one line of the extending frame for choosing xtal-dataset
# pairs

  CreateMTZSelectLine line \
      "Choose a crystal and dataset" HKLIN \
      "$counter: xtal" RUN_XTAL \
      "dataset" RUN_DATASET
}

#-----------------------------------------------------------------------
proc phaker_view_mtz { arrayname } {
#-----------------------------------------------------------------------
# Throw up a new window displaying the MTZ hierarchy as a tree
  upvar #0 $arrayname array

  set mtzfile [GetFullFileName0 $arrayname HKLIN]
  if { ![file exists $mtzfile] } {
      return 0
  }

  # Open a top level window
  set w .phakermtzview
  if { ![OpenWindow $w "Select MTZ columns" "MTZ"] } {
      return 0
  }
  CreateFrame $w $arrayname -noscroll

  # Initialise the hierarchical viewer
  set fdisplay [hierarch_create "[subst $w].fdisp" "Select MTZ columns: $mtzfile"]
  pack $fdisplay -side top -fill x

  # Descend the hierarchy and build the graphical representation
  foreach xtal [GetMtzXtals $mtzfile] {
      set fxtal [hierarch_add_node $fdisplay $xtal -noupdate]
      foreach dataset [GetMtzDatasetsInXtal $mtzfile $xtal] {
	  set fdata [hierarch_add_node $fxtal $dataset -noupdate]
	  foreach col [GetMtzColsInDataset $mtzfile $xtal $dataset] {
	      set fcol [hierarch_add_node $fdata $col -nochildren -noupdate]
	  }
      }
  }

  # Update the display - forces drawing of connecting lines
  hierarch_update_display $fdisplay
  CreateButton $w dismiss Quit "destroy $w"
  CreateButton $w select OK "phaker_process_select $arrayname $w $fdisplay"
}

#-----------------------------------------------------------------------
proc phaker_process_select { arrayname w f } {
#-----------------------------------------------------------------------
# Process the selection returned from the hierarchical view
    set selection [return_selection $f]
    # Remove the viewer window
    destroy $w
    # Warn about empty selection
    if { $selection == "" } {
	WarningMessage "Nothing was selected"
	return
    }
    # Process the selection
    #
    # Number of selected items
    set message "There are [llength $selection] selected items:\n\n"
    #
    # Look at each selection in turn
    foreach selected $selection {
	foreach item $selected { append name "/$item" }
	append message "$name"
	# Check there are three items (xtal, dataset, column)
	if { [llength $selected] < 3 } {
	    append message " (incomplete selection)\n"
	} else {
	    append message " (column \"[lindex $selected 2]\")\n"
	}
    }
    WarningMessage $message
    return
}

#-------------------------------------------------------------------------
proc CreateMTZSelectLine { linein message filein xtal_label xtal \
			       dataset_label dataset args } {
#-------------------------------------------------------------------------
#d_sum Draw task interface line for selecting input MTZ xtals and datasets
#d_desc The MTZ  line allows the user to select a crystal and associated \
dataset, taken from the MTZ header. The dataset is one of those defined \
for the selected crystal, so when the user selects the crystal then \
the dataset menu is automatically updated.
#d_desc This was originally built on the CreateLabinLine command.
#d_arg linein Output Tk frame id for the line
#d_arg message Message line which appears if cursor over this line
#d_arg filein Name of element in the task interface array which contains \
the name of the input MTZ file
#d_arg xtal_label Text to appear in the task interface line
#d_arg xtal Name of the element in the task interface array which contains \
the name of the xtal
#d_arg dataset_label
#d_arg dataset
#d_opt0 -toggle_display toggle_var toggle_state toggle_hitlist
#d_opt1 Display this line in the task window dependendent on the value of \
toggle_var.  If it takes any of the values in the list toggle_hitlist then \
set the display status to toggle_state.  toggle_state can be 'open', the \
line is displayed, or  'hide', the line is not displayed.

  upvar $linein line
  set arrayname [ GetSystemVar frame_array ]
  upvar #0 $arrayname array

  set help_pointer mtz_xtal

  set fileline ""
  foreach w [GetWidget $arrayname $filein] {
      lappend fileline [winfo parent $w] }
  set file_status [CheckFileInput $arrayname $filein read ]

   set n 0; set nargs -1
   if { [info exists args] } { set nargs [expr [llength $args] - 1 ] }
   while { $n < $nargs } {
     set argument [lindex $args $n ]
     if { $argument == "-toggle_display" } {
       incr n; set toggle_var [lindex $args $n ]
       incr n; set toggle_state [lindex $args $n ]
       incr n; set toggle_hitlist [lindex $args $n ]
     } elseif { $argument == "-help" } {
        incr n; set  help_pointer [lindex $args $n ]
     }
     incr n
   }

  set index [GetSystemVar current_index_counter]

  # Create the line
  CreateLine line  \
    message $message \
    help $help_pointer \
    label $xtal_label \
    widget $xtal \
      -command "UpdateDatasetMenu $arrayname $filein $xtal $dataset" \
    label $dataset_label \
    widget $dataset

  # Bind the input file so that it updates when a file is selected
  add_file_command $fileline "SetXtalDataset $arrayname $filein \
                [Indxv $xtal $index] \
                [Indxv $dataset $index] "

  # If a file has already been selected then do the update now
  if { $file_status } {
    eval SetXtalDataset $arrayname $filein [Indxv $xtal $index] \
	[Indxv $dataset $index]
  }

  # Enabling toggling of the line
  if { [info exists toggle_var ] } {
    toggle_frame_display $line $toggle_var $toggle_state $toggle_hitlist
  }
  return
}

# INTERNAL FUNCTION
# Called from within CreateMTZSelectLine when a file is selected
# It will update the xtal and dataset menus
#-------------------------------------------------------------------------
proc SetXtalDataset { arrayname filename xtal dataset args } {
#-------------------------------------------------------------------------
  upvar #0 $arrayname array
  global configure

  # Initialise
  set status 1
  set filestatus 1

  # Check args
    set allelements 0
    foreach arg $args {
	if {[regexp -- "-setallelements" $arg]} {
	    set allelements 1
	}
    }

  # Update the xtal menu
  set element $xtal
  GetIndx $element root c1 c2
  if { $c1 > 0 } {
    if { $allelements } {
      set elements_list [GetIndexedElements $arrayname $root $c1]
      if {[llength $elements_list] == 0 } { return }
    } else {
      set elements_list $element
    }
  } else {
    set elements_list $element
  }
  set xtal_list [GetMtzXtals [GetFullFileName0 $arrayname $filename]]
  lappend xtal_list "Unassigned"

  foreach ele $elements_list {

    # There may be more than one occurance of this widget
    # in the window - update all occurances
    set field_list [GetWidget $arrayname $ele ]

    foreach field $field_list {
      initialise_menu $field
      foreach xtalnam $xtal_list {
        $field.m add command -label "$xtalnam"  \
          -font $configure(FONT_REGULAR) \
	  -command "UpdateDatasetMenu \"$xtalnam\" $arrayname $filename $ele $dataset"
      }
    }
    # Update the current xtal if it is not in the list
    if { [lsearch $xtal_list $array($ele)] < 0 } {
	set xtalnam [lindex $xtal_list 0]
    } else {
	set xtalname $array($ele)
    }
    # Update the corresponding dataset
    UpdateDatasetMenu $xtalnam $arrayname $filename $ele $dataset
  }
  return
}

#INTERNAL FUNCTION
#-------------------------------------------------------------------------
proc UpdateDatasetMenu { xtalnam arrayname filename xtal dataset } {
#-------------------------------------------------------------------------
# Update the dataset menu widget corresponding to an xtal widget
# Invoked whenever the selected xtal is changed
  upvar #0 $arrayname array
  global configure

  # Initialise
  set status 1
  set filestatus 1
  set allelements 0

  # Update the xtal widget
  set array($xtal) $xtalnam

  # Update the dataset menu
  set element $dataset
  GetIndx $element root c1 c2
  if { $c1 > 0 } {
    if { $allelements } {
      set elements_list [GetIndexedElements $arrayname $root $c1]
      if {[llength $elements_list] == 0 } { return }
    } else {
      set elements_list $element
    }
  } else {
    set elements_list $element
  }
  if { [regexp -- ^Unassigned$ $xtalnam] } {
      set dataset_list {}
  } else {
      set dataset_list [GetMtzDatasetsInXtal \
	      [GetFullFileName0 $arrayname $filename] $xtalnam]
  } 
  lappend dataset_list "Unassigned"

  foreach ele $elements_list {

    # There may be more than one occurance of this widget
    # in the window - update all occurances
    set field_list [GetWidget $arrayname $ele ]

    foreach field $field_list {
      initialise_menu $field
      foreach datasetnam $dataset_list {
          $field.m add command -label "$datasetnam"  \
          -font $configure(FONT_REGULAR) \
          -command "SetDataset \"$datasetnam\" $arrayname $ele" 
      }
    }
    # Update the current dataset if it is not in the list
    if { [lsearch $dataset_list $array($ele)] < 0 } {
	set datasetnam [lindex $dataset_list 0]
    } else {
        set datasetnam $array($ele)
    }
    SetDataset $datasetnam $arrayname $ele
  }
  return
}

#-------------------------------------------------------------------------
proc SetDataset { datasetnam arrayname dataset } {
#-------------------------------------------------------------------------
  upvar #0 $arrayname array

  # Update the xtal widget
  set array($dataset) $datasetnam
  return
}
