# CCP4Interface  hierarch_utils.tcl
#
#     This code is distributed under the terms and conditions of the
#     CCP4 licence agreement as `Part ii)' software.  See the conditions
#     in the CCP4 manual for a copyright statement.
#
# Commands and utilities for handling tree-based hierarchical viewer
#
# Peter Briggs CCP4 September 2002
##################################################################

#CCP4i_cvs_Id $Id: hierarch_utils.tcl,v 1.13 2006/10/10 14:58:03 pjx Exp $
#
#d_index_title Hierarchical Viewer Utilities (utils/hierarch.tcl)
#d_intro_title Hierarchical Viewer Utilities
#
#d_intro  A set of utilities for creating, querying and editing a tree \
based visual data structure
#
# This code is an attempt to solve the problem of displaying
# hierarchical data by mapping it onto the Tk widget hierarchy.
# As much as possible the display is the data structure.
#
# Description of node structure and naming:
# -----------------------------------------
#
# (parent).fnode_$i
# (parent).fnode_$i.name
# (parent).fnode_$i.control
# (parent).fnode_$i.fchild
# (parent).fnode_$i.data (not displayed)
# (parent).fnode_$i.editname (used for editing, usually not displayed)
#
# The fchild is a container for the subtree below the
# current node. It has the following members
#
# (parent).fnode_$i.fchild.connect
# (parent).fnode_$i.fchild.fsubnodes
#
# The subnodes frame wraps the nodes of the subtree, i.e.
#
# (parent).fnode_$i.fchild.fsubnodes.fnode_$j
#
# As a rule (not yet strictly enforced) operations should only
# require the passing of the node widget path - the component
# names being generated internally.
#
# The root node name is defined by the application. It has a
# special label (parent).root widget (never displayed) to
# distinguish it from other nodes on the tree branches.
#
# Usage:
# Call hierarch_create to make a new tree frame
# Call hierarch_add_node to add nodes
# Call hierarch_update_display to force redrawing of connectivity
#      lines (nb this can be very slow at present)
# Call return_selection to get back a list of the names of selected
#      nodes

# Create the widget assembly for the hierarchical view
#-----------------------------------------------------------------------
proc hierarch_create { win { title "" } } {
#-----------------------------------------------------------------------
    # Create the hull frame to contain everything else
    # This also needs to create a "root" node from which
    # all others descend

    frame $win -class Hierarch

    # Create a root label which is never displayed
    label $win.root -text "ROOT"

    if { $title != "" } {
	label $win.title -text $title
	pack $win.title -side top -anchor w
    }
    frame $win.border -borderwidth 2 -relief groove
    pack $win.border -expand yes -fill both
    return $win
}

#-----------------------------------------------------------------------
proc hierarch_add_node { wnode name args } {
#-----------------------------------------------------------------------
#d_sum Add a new child node to the hierarchy
#d_desc
#d_arg wnode The widget pathname of the parent node
#d_arg name  Text to be displayed
#d_opt0 -nochildren
#d_opt1 Do not add placeholders for children of the new node i.e. made \
it a "leaf" branch of the tree.
#d_opt0 -noupdate
#d_opt1 Do not update the display (mainly the connectivity diagrams, since \
updating the display can be time consuming).
#d_opt0 -noexpand
#d_opt1 Do not make the child node display expandable/collapsable
#d_opt0 -callback cmd
#d_opt1 Specify a callback function to be invoked when the user does a \
right mouse click on the node
#d_opt0 -mouse_enter cmd
#d_opt1 Specify a callback function to be invoked when the user moves the \
cursor into the node label
#d_opt0 -mouse_leave cmd
#d_opt1 Specify a callback function to be invoked when the user moves the \
cursor out of the node label
#d_opt0 -data key value
#d_opt1 Store a data item as a key-value pair for this node. Multiple \
-data arguments can be used
#d_opt0 -mode <mode>
#d_opt1 Not implemented: <mode> can be "expanded" (default) or collapsed, \
and controls whether the subnodes of the new node are initialised as \
expanded or collapsed branches.
    global env

    # Arbitrary maxium number of nodes
    set maxnodes 1000

    # Initialise default settings
    set children 1
    set update_display 1
    set expand_on 1
    set callbackCmd ""
    set mouseenterCmd ""
    set mouseleaveCmd ""
    set data_list {}

    # Loop over args
    set n 0
    while { $n < [llength $args] } {
        set arg [lindex $args $n]
	# No children
	if { [regexp -- "-nochildren" $arg] } {
	    set children 0
	}
	# Don't update display automatically
	if { [regexp -- "-noupdate" $arg] } {
	    set update_display 0
	}
	# Don't make the child nodes expandable
	if { [regexp -- "-noexpand" $arg] } {
	    set expand_on 0
	}
	# Specify a callback function
	if { [regexp -- "-callback" $arg] } {
	    incr n
	    set callbackCmd [lindex $args $n]
	}
	# Specify a mouse_enter callback function
	if { [regexp -- "-mouse_enter" $arg] } {
	    incr n
	    set mouseenterCmd [lindex $args $n]
	}
	# Specify a mouse_leave callback function
	if { [regexp -- "-mouse_leave" $arg] } {
	    incr n
	    set mouseleaveCmd [lindex $args $n]
	}
	# Store key-value pair
	if { [regexp -- "-data" $arg] } {
	    incr n; set key [lindex $args $n]
	    incr n; set value [lindex $args $n]
	    lappend data_list $key $value
	}
	# Next argument
	incr n
    }

    # Determine name of subframe into which new node will be
    # added
    # The root window is a special case
    if { [winfo exists $wnode.root] } {
	set wparent $wnode
    } else {
	set wparent "$wnode.fchild.fsubnodes"
    }
    if { ![winfo exists $wparent] } {
	# Parent doesn't exist so cannot add the node
	return ""
    }

    # Fetch a name for the new subnode
    set i 1
    while { $i < $maxnodes && [winfo exists "$wparent.fnode_$i"] } {
	incr i
    }
    if { $i == $maxnodes } {
	# Exceeded arbitrary maximum number of nodes
	puts "hierarch_add: too many child nodes for $wnode"
	return ""
    }
    # Frame to contain the child node (fnode)
    set fnode [frame "$wparent.fnode_$i"]

    # Label with the name
    set lname [label "$fnode.name" -text $name \
	    -width 15 -borderwidth 2 -relief raised]

    # Label with the data items
    # This is a hidden label and is never displayed - it is
    # just used to store the associated data
    set ldata [label "$fnode.data" -text $data_list]

    # Entry widget for editing the name
    # This is hidden by default
    set tedit [entry "$fnode.editname" -width 15]

    # Pack permanently displayed items
    pack $fnode -side top -fill x
    pack $lname -side left -anchor ne

    # Add subframes for the descendents of this node
    if { $children } {

	# Control button for expansion/collapse of future subnodes
	if { $expand_on } {
	    set lcontrol [label "[subst $fnode].control" \
			      -bitmap @[file join $env(CCP4I_TOP) icons hier_collapse.xbm]]
	    # Pack the control button
	    pack $lcontrol -side left -anchor ne
	}

	# Add subframes for the descendents of this node
	set fchild [frame "[subst $fnode].fchild"]
	set fsubnodes [frame "[subst $fchild].fsubnodes"]

	# Add a canvas for displaying the connectivity of the descendents
	# NB width is also specified in proc hierarch_update_connectivity
	set width 20
	set connect [canvas "[subst $fchild].connect" \
		-height 0 -width $width -borderwidth 0]

	# Pack remaining child items
	pack $fchild -side left -anchor ne
	pack $connect $fsubnodes -side left -anchor nw
    }

    # Add bindings
    # 1. Bind command to select an item
    bind $lname <ButtonPress-1> "select_item $lname $fnode"
    # 2. Bind command to collapse the display of subnodes (only if
    #    subnodes are allowed and display is collapsible)
    if { $children && $expand_on } {
	bind $lcontrol <ButtonPress-1> "collapse_hierarchy $fnode"
    }
    # 3. Bind command to callback function for right mouse click
    if { $callbackCmd != "" } {
        bind $lname <ButtonPress-3> "$callbackCmd $fnode %X %Y"
    }
    # 4. Bind mouseover callback functions for node label
    if { $mouseenterCmd != "" } {
	bind $lname <Enter> "$mouseenterCmd $fnode %X %Y"
    }
    if { $mouseleaveCmd != "" } {
	bind $lname <Leave> "$mouseleaveCmd $fnode %X %Y"
    }

    # Update the connectivity lines
    if { $update_display } {
	hierarch_update_connectivity $wnode
    }

    # Return the path for the node
    return $fnode
}

# Copy a node
#-----------------------------------------------------------------------
proc hierarch_copy_node { wnode wtarget args } {
#-----------------------------------------------------------------------
    # Make a copy of wnode under wtarget
    # -all: recursively also copy all descendents of wnode
    # -name: specify a new name for the copy

    # Initialise defaults
    set copy_all 0
    set alt_name ""

    # Process arguments
    set n 0
    while { $n < [llength $args] } {
        set arg [lindex $args $n]
	# Copy all descendents
	if { [regexp -- "-all" $arg] } {
	    set copy_all 1
	}
	# Specify a new name for the copy
	if { [regexp -- "-name" $arg] } {
	    incr n
	    set alt_name [lindex $args $n]
	}
	# Next argument
	incr n
    }

    # Names of internal components of the source node
    set lname $wnode.name
    set lcontrol $wnode.control
    set ldata $wnode.data
    set fchild $wnode.fchild

    # Create a new node
    # Get the node name
    if { "$alt_name" == "" } {
	set name [hierarch_get_node_name $wnode]
    } else {
	set name $alt_name
    }
    set add_cmd "hierarch_add_node $wtarget $name"

    # Expandable
    if { ![winfo exists $wnode.control] } {
	append add_cmd " -noexpand"
    }

    # Children
    if { ![winfo exists $wnode.fchild] } {
	append add_cmd " -nochildren"
    }

    # Bindings
    # Right-mouse click
    set callbackCmd [bind $lname <ButtonPress-3>]
    if { "$callbackCmd" != "" } {
	# Process the binding to recover the actual command
	# Remove trailing "fnode %X %Y" arguments
	set n [expr [llength $callbackCmd] - 4]
	set callbackCmd [lrange $callbackCmd 0 $n]
	append add_cmd " -callback \{ $callbackCmd \}"
    }
    # Mouse-enter
    set mouseenterCmd [bind $lname <Enter>]
    if { "$mouseenterCmd" != "" } {
	# Process the binding to recover the actual command
	# Remove trailing "fnode %X %Y" arguments
	set n [expr [llength $mouseenterCmd] - 4]
	set mouseenterCmd [lrange $mouseenterCmd 0 $n]
	append add_cmd " -mouse_enter \{ $mouseenterCmd \}"
    }
    # Mouse-leave
    set mouseleaveCmd [bind $lname <Leave>]
    if { "$mouseleaveCmd" != "" } {
	# Process the binding to recover the actual command
	# Remove trailing "fnode %X %Y" arguments
	set n [expr [llength $mouseleaveCmd] - 4]
	set mouseleaveCmd [lrange $mouseleaveCmd 0 $n]
	append add_cmd " -mouse_leave \{ $mouseleaveCmd \}"
    }
    
    # Create the copy node
    set fnode [eval $add_cmd]

    # Copy the associated data
    foreach item [hierarch_get_node_data $wnode] {
	hierarch_set_node_data_item $fnode [lindex $item 0] [lindex $item 1]
    }

    # Finally: copy descendent nodes if requested
    if { $copy_all } {
	foreach wchild [hierarch_get_children $wnode] {
	    hierarch_copy_node $wchild $fnode -all
	}
    }

    # Finished
    return $fnode
}

# Delete a node and all its descendents
#-----------------------------------------------------------------------
proc hierarch_delete_node { wnode } {
#-----------------------------------------------------------------------
    destroy $wnode
    return
}

# Delete all nodes i.e. clear the display
#-----------------------------------------------------------------------
proc hierarch_delete_all_nodes { win } {
#-----------------------------------------------------------------------
    set wnodes [hierarch_get_children $win]
    foreach w $wnodes {
	hierarch_delete_node $w
    }
}

#-----------------------------------------------------------------------
proc hierarch_update_display { win } {
#-----------------------------------------------------------------------
#d_sum Force update of the whole tree display
#d_desc Invoke this to force (re)drawing of the connectivity lines
#d_arg win The Tk widget path name for the top level node in the tree
    # Get the children of this window
    set wchildren [hierarch_get_children $win]
    if { [llength $wchildren] > 0 } {
	# Call recursively until no more children are found
	foreach wchild $wchildren {
	    hierarch_update_display $wchild
	}
    }
    hierarch_update_connectivity $win
    return
}

## DISPLAY OF CONNECTIVITY LINES ########################################

# Create or update the connectivity lines for the children of
# wnode
#-----------------------------------------------------------------------
proc hierarch_update_connectivity { wnode } {
#-----------------------------------------------------------------------
    # Default line width
    set linewidth 2

    # Default width and half-width (pixels) of connectivity diagram
    # This should probably be configurable
    set width 20
    set hwidth [expr $width / 2]

    # Update the display to make sure everything is displayed
    update idletasks

    # Generate the name of the connectivity display widget
    set connect "[subst $wnode].fchild.connect"

    # Frame holding all the nodes
    set fsubnodes "[subst $wnode].fchild.fsubnodes"
    if { ![winfo exists $fsubnodes] } {
	# If there are no children then return
	return ""
    }

    # Get the current height
    set height [winfo height $fsubnodes]

    # Generate the name of the connectivity display widget
    set connect "[subst $wnode].fchild.connect"
    # Delete the current contents of the canvas displaying the
    # connectivity, and resize
    $connect delete all
    $connect configure -height $height

    # Draw the lines
    set i 0
    foreach fnode [hierarch_get_children $wnode] {
	# First subnode (=first line in the connectivity diagram)
	# is a special case
	if { $i == 0 } {
	    # Get the height of the name label
	    set lheight [expr [winfo height [subst $fnode].name] / 2]
	    # Line is full width
	    $connect create line 0 $lheight $width $lheight -width $linewidth
	    set ypos0 $lheight
	    incr i
	} else {
	    # Get the y position of the container
	    set ypos1 [expr [winfo y $fnode] + $lheight]
	    # Line is an elbow
	    $connect create line $hwidth $ypos0 $hwidth $ypos1 $width $ypos1 \
		     -width $linewidth
	    # Prepare for next line
	    set ypos0 $ypos1
	}
    }

    # Propagate the update up the tree
    set wparent [hierarch_get_parent $wnode]
    if { $wparent != "" } {
	hierarch_update_connectivity $wparent
    }
    return
}

#-----------------------------------------------------------------------
proc collapse_hierarchy { wnode } {
#-----------------------------------------------------------------------
# Collapse the display of the subelements by unpacking them
    # Fetch names of widgets for display of icon and children
    global env
    set wicon "$wnode.control"
    set wdata "$wnode.fchild"
    # Unpack the children
    pack forget $wdata
    # Change the icon displayed
    $wicon configure -bitmap @[file join $env(CCP4I_TOP) icons hier_expand.xbm]
    # Update the binding to expand the hierarchy again
    bind $wicon <ButtonPress-1> "expand_hierarchy $wnode"
    # Update the connecting lines
    hierarch_update_connectivity $wnode
}

#-----------------------------------------------------------------------
proc expand_hierarchy { wnode } {
#-----------------------------------------------------------------------
# Expand the display of the subelements by repacking them
    global env
    # Fetch names of widgets for display of icon and children
    set wicon "$wnode.control"
    set wdata "$wnode.fchild"
    # Repack the children
    pack $wdata -after $wicon -side left -anchor nw
    # Change the icon displayed
    $wicon configure \
	-bitmap @[file join $env(CCP4I_TOP) icons hier_collapse.xbm]
    # Update the binding to expand the hierarchy again
    bind $wicon <ButtonPress-1> "collapse_hierarchy $wnode"
    # Update the connecting lines
    hierarch_update_connectivity $wnode
}

## EDIT THE NODE NAME ###################################################

#-----------------------------------------------------------------------
proc hierarch_edit_node_name { wnode { check_proc "" } } {
#-----------------------------------------------------------------------
    # Swap the label displaying the name for an entry widget that
    # the user can edit
    # Optionally invoke a user-specified function "check_proc"
    # when editing is finished (completion of the edit process is
    # performed by calling hierarch_edit_node_finish)
    set lname $wnode.name
    set tedit $wnode.editname
    if { [winfo exists $tedit] && [winfo exists $lname] } {
	# Load the name to be edited
	$tedit delete 0 end
	$tedit insert 0 [hierarch_get_node_name $wnode]
	# Hide the label and show the entry box
	set wbefore [lindex [pack slaves $wnode] 0]
	pack $tedit -side left -anchor ne -before $wbefore
	pack forget $lname
	# Give it focus
	focus $tedit
	# Add bindings
	bind $tedit <KeyPress-Return> "hierarch_edit_node_finish $wnode $check_proc"
	bind $tedit <FocusOut> "hierarch_edit_node_finish $wnode $check_proc"
    }
}

#-----------------------------------------------------------------------
proc hierarch_edit_node_finish { wnode { check_proc "" } } {
#-----------------------------------------------------------------------
    # Perform completion of an node name editing operation
    # If defined the function "check_proc" is invoked, and
    # the change is finalised if this returns 1; if zero
    # then cancel the change.
    # If -1 then return to the editing mode
    set lname $wnode.name
    set tedit $wnode.editname
    if { [winfo exists $tedit] && [winfo exists $lname] } {
	set name [$tedit get]
	# Check whether the name has actually changed
	if { "$name" == [hierarch_get_node_name $wnode] } {
	    hierarch_hide_editname $wnode
	    return
	}
	if { $check_proc != "" } {
	    set status [eval $check_proc $wnode { $name } ]
	    if { $status > 0 } {
		# Ok to use this name
		hierarch_set_node_name $wnode "$name"
	    } elseif { $status < 0 } {
		# Return to editing mode
		return
	    }
	} else {
	    # Just accept the change
	    hierarch_set_node_name $wnode "$name"
	}
	# Remove the bindings on the entry widget
	foreach tag [bind $tedit] {
	    bind $tedit $tag {}
	}
	# Hide the editing widget and restore the label
	hierarch_hide_editname $wnode
    }
}

#-----------------------------------------------------------------------
proc hierarch_hide_editname { wnode } {
#-----------------------------------------------------------------------
    # Hide the editname widget and show the label displaying the name
    set lname $wnode.name
    set tedit $wnode.editname
    if { [winfo exists $tedit] && [winfo exists $lname] } {
	# Determine the widget to insert before
	set wbefore [lindex [pack slaves $wnode] 0]
	pack $lname -side left -anchor ne -before $wbefore
	pack forget $tedit
    }
}

## QUERY/MODIFY ASSOCIATED DATA #########################################

#-----------------------------------------------------------------------
proc hierarch_get_node_data { wnode } {
#-----------------------------------------------------------------------
# Return all the stored data associated with the named node
    ##puts "\thierarch_get_node_data: $wnode"
    if { [winfo exists $wnode] } {
	set ldata $wnode.data
	if { [winfo exists $ldata] } {
	    ##puts "\thierarch_get_node_data: [$ldata cget -text]"
	    return [$ldata cget -text]
	}
    }
    ##puts "\thierarch_get_node_data: return blank"
    return {}
}

#-----------------------------------------------------------------------
proc hierarch_get_node_data_item { wnode key } {
#-----------------------------------------------------------------------
# Return the value of the specified data item for the named node
    ##puts "hierarch_get_node_data_item: requested data item is $key"
    set data_list [hierarch_get_node_data $wnode]
    ##puts "hierarch_get_node_data_item: $data_list"
    foreach data_item $data_list {
	##puts "*** Item: $data_item"
	if { [llength $data_item] == 2 } {
	    ##puts "     (contains 2 elements)"
	    if { [lindex $data_item 0] == "$key" } {
		##puts "Returning value [lindex $data_item 1]"
		return [lindex $data_item 1]
	    }
	}
    }
    # Failed to match the key
    ##puts "No match - returning blank"
    return {}
}

#-----------------------------------------------------------------------
proc hierarch_append_node_data { wnode key value } {
#-----------------------------------------------------------------------
# Append a key-value data pair to the named node
    set ldata $wnode.data
    set data_list [hierarch_get_node_data $wnode]
    lappend data_list [list $key $value]
    $ldata configure -text $data_list
    return [$ldata cget -text]
}

#-----------------------------------------------------------------------
proc hierarch_set_node_data_item { wnode key value } {
#-----------------------------------------------------------------------
# (Re)set the value of the specified data item for the named node
    set ldata $wnode.data
    ##puts "Data is $key $value"
    set data_list [hierarch_get_node_data $wnode]
    ##puts "Existing data list: $data_list"
    set new_data_list {}
    set need_append 1
    foreach data_item $data_list {
	if { [llength $data_item] == 2 } {
	    if { [lindex $data_item 0] == "$key" } {
		lappend new_data_list [list $key $value]
		set need_append 0
	    } else {
		lappend new_data_list $data_item
	    }
	}
    }
    if { $need_append } {
	# The key wasn't found in the existing data list
	# so it's a new data item
	lappend new_data_list [list $key $value]
    }
    ##puts "new_data_list: $new_data_list"
    $ldata configure -text $new_data_list
    ##puts "Data set to [$ldata cget -text]"
    return [$ldata cget -text]
}

## SELECTION UTILITIES ##################################################

#-----------------------------------------------------------------------
proc select_item { witem wframe } {
#-----------------------------------------------------------------------
# Select an item by changing its display properties
# Selection of an item automatically selects all child items (cascade).
# It also selects the parent items above it back to the root node
# (propagation). This is different to deselection behaviour.

    # Initialise colours of selected items
    # Ultimately these should be configurable
    set select_bg black
    set select_fg white

    $witem configure -background $select_bg -foreground $select_fg
    # Update binding so it can be deselected
    bind $witem <ButtonPress-1> "deselect_item $witem $wframe"
    # Cascade selection down the hierarchy
    cascade_select $wframe select_item
    # Propagate up the hierarchy
    propagate_selection $wframe select_item
}

#-----------------------------------------------------------------------
proc deselect_item { witem wframe } {
#-----------------------------------------------------------------------
# Deselect an item by changing its display properties
# Deselection of an item automatically deselects all child items
# (cascade). It doesn't affect the selection of any of its parent items.
# This is different to selection behaviour.

    # Initialise colours of selected items
    # Ultimately these should be configurable
    set select_bg grey
    set select_fg black

    $witem configure -background $select_bg -foreground $select_fg
    # Update binding so it can be reselected
    bind $witem <ButtonPress-1> "select_item $witem $wframe"
    # Cascade selection down the hierarchy
    cascade_select $wframe deselect_item
}

#-----------------------------------------------------------------------
proc propagate_selection { wnode selectProc } {
#-----------------------------------------------------------------------
# Propagate a selection up the directory tree
# This is quite tricky because it means travelling up
# the tree structure
    # Identify parent node
    set path [hierarch_get_parent $wnode]
    # If there is a parent then propagate up the tree
    if { $path != "" } {
	# Generate the path for the label
	set witem "[subst $path].name"
	# Update the selection
	# This is clumsy as it duplicates the code
	# in the (de)select procedures, but does the
	# job for now
	if { [regexp "deselect_item" $selectProc] } {
	    # Deselect
	    $witem configure -background grey -foreground black
	    bind $witem <ButtonPress-1> "select_item $witem $path"
	} elseif { [regexp "select_item" $selectProc] } {
	    # Select
	    $witem configure -background black -foreground white
	    bind $witem <ButtonPress-1> "deselect_item $witem $path"
	}
	# Propagate up to the next level
	propagate_selection $path $selectProc
	# Placeholder: if there are commands associated with
	# selection and deselection then they should be executed
	# here
    }
    return
}

#-----------------------------------------------------------------------
proc cascade_select { wnode selectProc } {
#-----------------------------------------------------------------------
# Cascade a selection event down the hierarchy
    foreach pframe [hierarch_get_children $wnode] {
	# ... update the selection too
	eval "$selectProc [subst $pframe].name $pframe"
    }
    # Placeholder: if there are commands associated with
    # selection and deselection then they should be executed
    # here
}

#-----------------------------------------------------------------------
proc return_selection { wnode } {
#-----------------------------------------------------------------------
#d_sum Return a list of selected branches in the hierarchy
#d_desc This returns a list of lists - each sublist is a list of the \
names of selected nodes below the root.
#d_arg wnode The path of the root node

    # Get the paths of all the selected nodes in the hierarchy
    set selected_nodes [return_selected_nodes $wnode]
    ##puts "selected nodes: $selected_nodes"

    # Build a list of the longest unique path names
    # This will form the final selection
    set unique_nodes {}
    for { set i 0 } { $i < [llength $selected_nodes] } { incr i } {
	# For each selected node path:
	# see how many times it occurs in the list as a
	# "subpath" of other nodes (including itself)
	set node [lindex $selected_nodes $i]
	##puts "node: $node"
	set tmp $selected_nodes
	set nnode 0
	while { [set k [lsearch -regexp $tmp "^$node"]] > -1 } {
	    incr nnode
	    set tmp [lreplace $tmp $k $k]
	    ##puts ">> at position $k (occurance $nnode)"
	}
	# Each node should occur at least once so this is 
	# a serious error
	if { $nnode == 0 } {
	    puts "SERIOUS INTERNAL ERROR in return_selection (no occurances of $node)"
	    return {}
	}
	# Node paths that appear exactly once are the
	# longest unique paths
	# These are the ones that we wish to store
	if { $nnode == 1 } {
	    lappend unique_nodes $node
	}
    }
    ##puts ">> unique nodes: $unique_nodes"
    # Build another list with the name of each unique node
    # This forms the selection to be returned
    set selection {}
    foreach node $unique_nodes {
	lappend selection [hierarch_get_node_full_name $node]
    }
    ##puts "return_selection: $selection"
    return $selection
}

#-----------------------------------------------------------------------
proc return_selected_nodes { wnode { selected {} } } {
#-----------------------------------------------------------------------
#d_sum Return a list of all selected node names
#d_sum This returns an unsorted list of all nodes that are selected.
#d_arg wnode The node to generate the selection from
#d_arg selected Previous selection to append the node list to (used \
internally)

    # First check if we are at the root node
    # Root is unusual since it cannot be selected
    if { [winfo exists $wnode.root] } {
	##puts "$wnode: is root"
	set selection ""
    } else {
	##puts "$wnode"
	# Check if this (non-root) node is selected
	if { [hierarch_is_selected_node $wnode] } {
	    # Add to the list of selected nodes
	    ##puts ">>>> $wnode: is selected"
	    lappend selected $wnode
	}
    }
    # Descend each branch gathering selections
    ##puts ">> $wnode: looping over children"
    ##puts ">> children are [hierarch_get_children $wnode]"
    foreach child [hierarch_get_children $wnode] {
	##puts ">>>> Now examining $child"
	set selected [return_selected_nodes $child $selected]
    }
    # Finished - return
    ##puts ">> Returning selection \"$selected\""
    ##puts "******** finished *******"
    return $selected
}

#-----------------------------------------------------------------------
proc hierarch_get_selected_children { wnode } {
#-----------------------------------------------------------------------
#d_sum Return list of selected child nodes
    if { [llength [set children [hierarch_get_children $wnode]]] < 1 } {
	# No children
	return {}
    }
    set selected_children {}
    foreach child $children {
	if { [hierarch_is_selected_node $child] } {
	    lappend selected_children $child
	}
    }
    return $selected_children
}

#-----------------------------------------------------------------------
proc hierarch_is_selected_node { wnode } {
#-----------------------------------------------------------------------
#d_sum Return 1 if node is selected, 0 if not
    # Node is considered to be selected if background colour is
    # same as that for selection
    # NB This should ultimately be a global parameter, or else a special
    # unseen widget should be used as a flag
    set select_bg black
    set wname $wnode.name
    if { [winfo exists $wname] } {
	##puts "$wname: exists, background is \"[$wname configure -background]\""
	if { [regexp -- "$select_bg\$" [$wname configure -background]] } {
	    return 1
	}
    }
    return 0
}

#-----------------------------------------------------------------------
proc hierarch_find_node { wnode name } {
#-----------------------------------------------------------------------
#d_sum Find a subnode of the specified node with the given name
    set wchildren [hierarch_get_children $wnode]
    # Does the name of this node match?
    if { [hierarch_get_node_name $wnode] == $name } {
	set wnodes $wnode
    } else {
	set wnodes {}
    }
    # Now call this function recursively
    foreach w $wchildren {
	set wnodes [concat $wnodes [hierarch_find_node $w $name]]
    }
    return $wnodes
}

#-----------------------------------------------------------------------
proc hierarch_node_level { wnode } {
#-----------------------------------------------------------------------
#d_sum Return the level of the node
#d_desc Given a node this function returns the "depth" of the node
    # in terms of its distance from the root node
    if { ![winfo exists $wnode] } {
	return -1
    }
    # Check if we are at the root node
    if { [winfo exists $wnode.root] } {
	return 0
    }
    set level 0
    set eles [split $wnode "."]
    foreach ele $eles {
	if { [regexp -- "^fnode_\[0-9\]+\$" $ele] } {
	    incr level
	}
    }
    return $level
}

#-----------------------------------------------------------------------
proc hierarch_get_node_name { wnode } {
#-----------------------------------------------------------------------
#d_sum Return the name (text) associated with the specified node
    set wname $wnode.name
    if { [winfo exists $wname] } {
	return [lindex [$wname cget -text] end]
    }
    return ""
}

#-----------------------------------------------------------------------
proc hierarch_set_node_name { wnode name } {
#-----------------------------------------------------------------------
#d_sum Set the name (text) associated with the specified node
    set wname $wnode.name
    if { [winfo exists $wname] } {
	return [$wname configure -text "$name"]
    }
    return ""
}

#-----------------------------------------------------------------------
proc hierarch_get_node_full_name { wnode } {
#-----------------------------------------------------------------------
#d_sum Return the name (text) associated with the node plus
    # its parents all the way up to the root node
    set node $wnode
    set full_name [hierarch_get_node_name $wnode]
    while { [set node [hierarch_get_parent $node]] != "" } {
	##lappend full_name [hierarch_get_node_name $node]
	set full_name [linsert $full_name 0 [hierarch_get_node_name $node]]
    }
    return $full_name
}

## INTERNAL UTILITIES ###################################################
#
# Functions for finding parents and children

#-----------------------------------------------------------------------
proc hierarch_get_children { wnode } {
#-----------------------------------------------------------------------
# Return a list of child nodes of node wnode
    if { ![winfo exists $wnode] } {
	return ""
    }
    # If this is the root node then it is a special case
    if { [winfo exists $wnode.root] } {
	set widgets [winfo children $wnode]
    } else {
	# For non-root nodes
	set wsubnodes "$wnode.fchild.fsubnodes"
	if { ![winfo exists $wsubnodes] } {
	    return ""
	} else {
	    set widgets [winfo children $wsubnodes]
	}
    }
    # Only return node paths
    set wchildren {}
    foreach w $widgets {
	if { [regexp -- "\.fnode_\[0-9\]+\$" $w] } {
	    lappend wchildren $w
	}
    }
    return $wchildren
}

#-----------------------------------------------------------------------
proc hierarch_get_parent { wnode } {
#-----------------------------------------------------------------------
# Return the parent node of the child wnode
    if { ![winfo exists $wnode] } {
	return ""
    }
    # Check if the parent is root
    if { [hierarch_node_level $wnode] == 1 } {
	return ""
    }
    # Break widget name up into components
    set welements [split $wnode "."]

    # Remove the last three levels (subtract 4 because this the
    # indices start from zero) and regenerate name using join
    set n [expr [llength $welements] - 4]
    if { $n > 0 } {
	set wparent [join [lrange $welements 0 $n] "."]
	return $wparent
    }

    # No parent
    return ""
}

#-----------------------------------------------------------------------
proc hierarch_get_root { wnode } {
#-----------------------------------------------------------------------
#d_sum Return the root node of the tree
#d_desc
#d_arg wnode Any node in the tree
    if { ![winfo exists $wnode] } {
	return ""
    }
    # Assume that this is the root node
    set wroot $wnode
    set wparent [hierarch_get_parent $wnode]
    # Look for parents until we run out
    while { $wparent != "" } {
	set wroot $wparent
	set wparent [hierarch_get_parent $wparent]
    }
    # Return
    return $wroot
}
